from io import BytesIO
from ..General.GeneralClass import GeneralClass
from ..utils.getByte import getByte


class UseRuntimeIDPool(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'UseRuntimeIDPool'
        self.operationNumber: int = 31
        self.poolId: int = 0

    def Marshal(self, writer: BytesIO) -> None:
        writer.write(self.poolId.to_bytes(
            length=1, byteorder='big', signed=False))

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.poolId = getByte(buffer, 1)[0]



import nbtlib
import sys
from io import BytesIO
from struct import pack, unpack
from ..General.GeneralClass import GeneralClass
from ..utils.getByte import getByte


if sys.version_info.major >= 3 and sys.version_info.minor >= 10:
    from ..utils.marshalNBT import MarshalPythonNBTObjectToWriter
    from ..utils.unmarshalNBT import UnMarshalBufferToPythonNBTObject
else:
    from ..utils.marshalNBT_OldVersion import MarshalPythonNBTObjectToWriter
    from ..utils.unmarshalNBT_OldVersion import UnMarshalBufferToPythonNBTObject


class PlaceBlockWithNBTData(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'PlaceBlockWithNBTData'
        self.operationNumber: int = 41
        self.blockConstantStringID: int = 0
        self.blockStatesConstantStringID: int = 0
        self.blockNBT: nbtlib.tag.Compound = nbtlib.tag.Compound(
            {})  # type: ignore

    def Marshal(self, writer: BytesIO) -> None:
        writer.write(pack('>H', self.blockConstantStringID) + pack('>H',
                     self.blockStatesConstantStringID) + pack('>H', self.blockStatesConstantStringID))
        MarshalPythonNBTObjectToWriter(writer, self.blockNBT, '')

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.blockConstantStringID = unpack('>H', getByte(buffer, 2))[0]
        self.blockStatesConstantStringID = unpack('>H', getByte(buffer, 2))[0]
        getByte(buffer, 2)  # do not delete this line because it is correct
        self.blockNBT, _ = UnMarshalBufferToPythonNBTObject(  # type: ignore
            buffer)

    def Loads(self, jsonDict: dict) -> None:
        if 'operationDatas' in jsonDict:
            jsonDict = jsonDict['operationDatas']
            self.blockConstantStringID = jsonDict['blockConstantStringID'] if 'blockConstantStringID' in jsonDict else 0
            self.blockStatesConstantStringID = jsonDict[
                'blockStatesConstantStringID'] if 'blockStatesConstantStringID' in jsonDict else 0
            self.blockNBT = nbtlib.parse_nbt(
                jsonDict['blockNBT']) if 'blockNBT' in jsonDict else nbtlib.tag.Compound({})

    def Dumps(self) -> dict:
        return {
            'operationName': self.operationName,
            'operationNumber': self.operationNumber,
            'operationDatas': {
                'blockConstantStringID': self.blockConstantStringID,
                'blockStatesConstantStringID': self.blockStatesConstantStringID,
                'blockNBT': nbtlib.serialize_tag(self.blockNBT)
            }
        }



from io import BytesIO
from copy import deepcopy


class GeneralClass:
    """
    Any operation of the BDX file will inherit this class
    """

    def __init__(self) -> None:
        self.operationNumber: int
        self.operationName: str

    def Marshal(self, writer: BytesIO) -> None:
        """
        Marshal Self@GeneralClass into the writer(io object)
        """
        ...

    def UnMarshal(self, buffer: BytesIO) -> None:
        """
        Unmarshal the buffer(io object) into Self@GeneralClass
        """
        ...

    def Loads(self, jsonDict: dict) -> None:
        """
        Load datas from jsonDict:dict
        """
        if 'operationNumber' in self.__dict__:
            jsonDict['operationNumber'] = self.operationNumber
            jsonDict['operationName'] = self.operationName
        if 'operationDatas' in self.__dict__:
            for i in self.__dict__['operationDatas']:
                if i in jsonDict:
                    self.__dict__['operationDatas'][i] = jsonDict[i]
        else:
            for i in self.__dict__:
                if i in jsonDict:
                    self.__dict__[i] = jsonDict[i]

    def Dumps(self) -> dict:
        """
        Convert Self@GeneralClass into the basic dictionary
        """
        if 'operationNumber' in self.__dict__:
            copyDict = deepcopy(self.__dict__)
            del copyDict['operationNumber']
            del copyDict['operationName']
            return {
                'operationNumber': self.operationNumber,
                'operationName': self.operationName,
                'operationDatas': copyDict
            }
        else:
            return self.__dict__



from io import BytesIO
from struct import pack, unpack
from ..General.GeneralClass import GeneralClass
from ..utils.getString import getByte, getString


class PlaceBlockWithBlockStatesDeprecated(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'PlaceBlockWithBlockStatesDeprecated'
        self.operationNumber: int = 13
        self.blockConstantStringID: int = 0
        self.blockStatesString: str = ''

    def Marshal(self, writer: BytesIO) -> None:
        writer.write(pack('>H', self.blockConstantStringID) +
                     self.blockStatesString.encode(encoding='utf-8') + b'\x00')

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.blockConstantStringID = unpack('>H', getByte(buffer, 2))[0]
        self.blockStatesString = getString(buffer)



from io import BytesIO
from struct import unpack
from .getByte import getByte
import nbtlib

endian: str = '<'


class getValueError(Exception):
    def __init__(self, errorOccurredPosition: int):
        Exception.__init__(
            self, f'failed to parse value, and the error occurred at position {errorOccurredPosition}')


class unexpectedError(Exception):
    def __init__(self, value: int):
        Exception.__init__(self, f'unexpected number {value}')


def getName(buffer: BytesIO) -> str:
    keyLength = unpack(f'{endian}H', getByte(buffer, 2))[0]
    return getByte(buffer, keyLength).decode(encoding='utf-8')


def getValue(buffer: BytesIO, valueType: int) -> ...:
    if valueType == 1:
        return nbtlib.tag.Byte(unpack(f'{endian}b', getByte(buffer, 1))[0])
    elif valueType == 2:
        return nbtlib.tag.Short(unpack(f'{endian}h', getByte(buffer, 2))[0])
    elif valueType == 3:
        return nbtlib.tag.Int(unpack(f'{endian}i', getByte(buffer, 4))[0])
    elif valueType == 4:
        return nbtlib.tag.Long(unpack(f'{endian}q', getByte(buffer, 8))[0])
    elif valueType == 5:
        return nbtlib.tag.Float(unpack(f'{endian}f', getByte(buffer, 4))[0])
    elif valueType == 6:
        return nbtlib.tag.Double(unpack(f'{endian}d', getByte(buffer, 8))[0])
    elif valueType == 7 or valueType == 11 or valueType == 12:
        return getArray(buffer, valueType)
    elif valueType == 8:
        return nbtlib.tag.String(getName(buffer))
    elif valueType == 9:
        return getList(buffer)
    elif valueType == 10:
        return getCompound(buffer)
    else:
        raise getValueError(buffer.seek(0, 1))


def getArray(buffer: BytesIO, valueType: int) -> ...:
    arrayLength = unpack(f'{endian}i', getByte(buffer, 4))[0]
    if valueType == 7:
        return nbtlib.tag.ByteArray([unpack(f'{endian}b', getByte(buffer, 1))[0] for _ in range(arrayLength)])
    elif valueType == 11:
        return nbtlib.tag.IntArray([unpack(f'{endian}i', getByte(buffer, 4))[0] for _ in range(arrayLength)])
    elif valueType == 12:
        return nbtlib.tag.LongArray([unpack(f'{endian}q', getByte(buffer, 8))[0] for _ in range(arrayLength)])
    else:
        raise unexpectedError(valueType)


def getList(buffer: BytesIO) -> nbtlib.tag.List:
    valueType = getByte(buffer, 1)[0]
    listLength = unpack(f'{endian}i', getByte(buffer, 4))[0]
    return nbtlib.tag.List([getValue(buffer, valueType) for _ in range(listLength)])


def getCompound(buffer: BytesIO) -> nbtlib.tag.Compound:
    result: dict = {}
    while True:
        nextBuffer = getByte(buffer, 1)
        if nextBuffer == b'\x00':
            return nbtlib.tag.Compound(result)
        # if meet TAG_End
        buffer.seek(-1, 1)
        # correct the pointer
        valueType = getByte(buffer, 1)[0]
        name = getName(buffer)
        result[name] = getValue(buffer, valueType)
        # set values


def UnMarshalBufferToPythonNBTObject(buffer: BytesIO) -> ...:
    valueType = getByte(buffer, 1)[0]
    name = getName(buffer)
    return getValue(buffer, valueType), name



from Crypto.Signature import PKCS1_v1_5
from Crypto.Hash.SHA256 import new
from Crypto.PublicKey import RSA
from io import BytesIO
from struct import pack, unpack
from .ErrorClassDefine import signatureError
from ..General.GeneralClass import GeneralClass
from ..utils.getByte import getByte


"""
import ecdsa

peer = ecdsa.SigningKey.generate(ecdsa.NIST384p)
verifyingKey = peer.get_verifying_key()
publicKey = verifyingKey.to_string().hex()

print(publicKey)
# publicKey(...)
"""
# Generate a new public key to send a auth request to the romote server


"""
The address of PhoenixBuilder Auth server is wss://api.fastbuilder.pro:2053
"""
# Address of PhoenixBuilder Auth server


"""
Golang Structure
type AuthRequest struct {
    Action         string `json:"action"`
    ServerCode     string `json:"serverCode"`
    ServerPassword string `json:"serverPassword"`
    Key            string `json:"publicKey"`
    FBToken        string
}

Python Dictionary
{
    'action': 'phoenix::login',
    'serverCode': ...,
    'serverPassword': ...,
    'publicKey': ...,
    'FBToken': ...
}
"""
# Send an auth request to the PhoenixBuilder Auth server
# Note: Must use GZIP to compress data when sending


"""
Python Dictionary
{
    'chainInfo': ...,
    'code': ...,
    'message': ...,
    'privateSigningKey': ...,
    'prove': ...
}
"""
# The response of the PhoenixBuilder Auth server when the request succeeds


class Signature(GeneralClass):
    """
    Note: `Signature` is not a operation
    """

    def __init__(self) -> None:
        """
        `prove: str` [You need to provide this while signing]
            Prove provided by the PhoenixBuilder Auth server.
                Note: You don't know what "Prove: str" is?
                If you go to the beginning of this file, 
                there's a comment there, then you will get 
                what you want. Same as follow
        `signer: str`
            Signer of this BDX file
        `legacySignBuffer: bytes`
            Binary signature datas on legacy signing method
        `signature: bytes`
            Binary signature datas
        `signedOrNeedToSign: bool` [You need to provide this while signing]
            If this file is signed, we will set it as True.
            Or, if you want to sign the changed BDX file,
            please set it as True
        `isLegacy: bool`
            It is true if the old signature method is used.
            The old signing method is deprecated,
            so we cannot support signing with this method
        `privateSigningKeyString: str` [You need to provide this while signing]
            The private key which used to sign BDX file
            provided by the PhoenixBuilder Auth server
        `BDXContentWithInsideHeader: io.BytesIO` [You need to provide this while only used to signing]
            The content of BDX file with inside file header
        """
        self.prove: str = ''
        self.signer: str = ''
        self.legacySignBuffer: bytes = b''
        self.signature: bytes = b''

        self.signedOrNeedToSign: bool = False
        self.isLegacy: bool = False

        self.privateSigningKeyString: str = ''
        self.BDXContentWithInsideHeader: BytesIO = BytesIO(b'')

    def verifyProve(self) -> str:
        """
        Verify self.prove and return signer's name
        """
        constantServerKey = RSA.import_key('-----BEGIN RSA PUBLIC KEY-----\nMIIBCgKCAQEAzOoZfky1sYQXkTXWuYqf7HZ+tDSLyyuYOvyqt/dO4xahyNqvXcL5\n1A+eNFhsk6S5u84RuwsUk7oeNDpg/I0hbiRuJwCxFPJKNxDdj5Q5P5O0NTLR0TAT\nNBP7AjX6+XtNB/J6cV3fPcduqBbN4NjkNZxP4I1lgbupIR2lMKU9lXEn58nFSqSZ\nvG4BZfYLKUiu89IHaZOG5wgyDwwQrejxqkLUftmXibUO4s4gf8qAiLp3ukeIPYRj\nwGhGNlUfdU0foCxf2QwAoBV2xREL8/Sx1AIvmoVUg1SqCiIVMvbBkDoFfkzPZCgC\nLtmbkmqZJnpoBVHcBhBdUYsfyM6QwtWBNQIDAQAB\n-----END RSA PUBLIC KEY-----')
        verifier = PKCS1_v1_5.new(constantServerKey)
        # get publick key and verifier
        # note: this public key is provided from PhoenixBuilder code library
        splitResult: list[str] = self.prove.split('::')
        if len(splitResult) != 2:
            raise signatureError(
                f'failed to parse prove datas; self.prove = "{self.prove}"')
        # split prove into list
        # example:
        # splitResult: list[str] = [rsaPublicKeyWithHolder: str, signature: str]
        # note: rsaPublicKey is used to sign the BDX file
        findResult = splitResult[0].find('-----END RSA PUBLIC KEY-----')
        if findResult == -1:
            raise signatureError(
                f'the prove provided has not been verified and may be invalid; self.prove = "{self.prove}"')
        signer = splitResult[0][findResult+30:]
        # get signer's name
        result = verifier.verify(
            new(splitResult[0].encode()),
            bytes.fromhex(splitResult[1])
        )
        if result == False:
            raise signatureError(
                f'the prove provided has not been verified and may be invalid; self.prove = "{self.prove}"')
        # verify the prove
        return signer
        # return

    def loadPrivateKey(self) -> RSA.RsaKey:
        """
        Load privateKeyString:str into RSA.RsaKey and return it
        """
        successStates: bool = True
        try:
            return RSA.import_key(self.privateSigningKeyString)
        except:
            successStates = False
        if successStates == False:
            raise signatureError(
                f'the privateSigningKeyString provided is invalid; self.privateSigningKeyString = "{self.privateSigningKeyString}"')

    def verifySignature(self) -> None:
        splitResult: list[str] = self.prove.split('|')
        if len(splitResult) != 2:
            raise signatureError(
                f'failed to parse prove datas; self.prove = "{self.prove}"')
        # split prove into list
        rsaPublicKey = RSA.import_key(splitResult[0])
        # load rsa public key
        verifier = PKCS1_v1_5.new(rsaPublicKey)
        # get verifier
        result = verifier.verify(
            new(self.BDXContentWithInsideHeader.getvalue()),
            self.signature
        )
        if result == False:
            raise signatureError(
                f'the signature provided has not been validated, and it may be invalid; self.signature.hex() = "{self.signature.hex()}"')
        # verify the signature

    def Marshal(self, writer: BytesIO) -> None:
        if self.signedOrNeedToSign == False or self.isLegacy == True:
            return
        # check if need to sign this file or this is a legacy method
        # note: legacy method is officially deprecated so we cannot support this
        if self.prove == '':
            raise signatureError('self.prove is not assigned')
        # check the states of self.prove
        if not 'privateSigningKeyString' in self.__dict__:
            raise signatureError('self.privateSigningKeyString is not existed')
        # check the states of self.privateSigningKeyString
        self.signer = self.verifyProve()
        # verification the prove
        newWriter: BytesIO = BytesIO(b'')
        newWriter.write(b'\x00\x8b')
        newWriter.write(pack('<H', len(self.prove)))
        newWriter.write(self.prove.encode(encoding='utf-8'))
        signatureDatas = PKCS1_v1_5.new(self.loadPrivateKey()).sign(
            new(self.BDXContentWithInsideHeader.getvalue()))
        newWriter.write(signatureDatas)
        self.signature = signatureDatas
        self.verifySignature()
        # get sign content and sync datas
        writer.write(b'X')
        writer.write(newWriter.getvalue())
        # write Terminate and sign content
        if newWriter.seek(0, 1) >= 255:
            writer.write(pack('>H', newWriter.seek(0, 1)))
            writer.write(b'\xff')
        else:
            writer.write(newWriter.seek(0, 1).to_bytes(
                length=1, byteorder='little', signed=False))
        # write the length of the sign content
        writer.write(b'\x5a')
        # write the number of this pseudo-operation

    def UnMarshal(self, buffer: BytesIO) -> None:
        nowSeek = buffer.seek(0, 1)
        buffer.seek(-1, 2)
        if getByte(buffer, 1) != b'\x5a':
            self.signedOrNeedToSign = False
            buffer.seek(nowSeek, 0)
            return
        else:
            self.signedOrNeedToSign = True
        # check if need to sign this file
        buffer.seek(-2, 2)
        if getByte(buffer, 1) == b'\xff':
            buffer.seek(-4, 2)
            signLength: int = unpack('>H', getByte(buffer, 2))[0]
            buffer.seek(-4-signLength, 2)
        else:
            buffer.seek(-2, 2)
            signLength: int = unpack('>B', getByte(buffer, 1))[0]
            buffer.seek(-2-signLength, 2)
        signContent = BytesIO(getByte(buffer, signLength))
        # get sign content ans sync datas
        if getByte(signContent, 2) != b'\x00\x8b':
            self.isLegacy = True
            self.legacySignBuffer = signContent.getvalue()
            return
        else:
            proveLength: int = unpack('<H', getByte(signContent, 2))[0]
            self.prove = getByte(
                signContent, proveLength).decode(encoding='utf-8')
            self.signer = self.verifyProve()
            self.signature = getByte(signContent, signLength-proveLength-4)
        # sync datas
        buffer.seek(nowSeek, 0)
        # revert the pointer

    def Loads(self, jsonDict: dict) -> None:
        self.prove = jsonDict['Prove'] if 'Prove' in jsonDict else ''
        self.prove = '' if self.prove == None else self.prove
        if 'Signer' in jsonDict:
            if jsonDict['Signer'] != None:
                self.signer = jsonDict['Signer']
        if 'Signature' in jsonDict:
            if jsonDict['Signature'] != None:
                self.signature = bytes.fromhex(jsonDict['Signature'])
        if 'LegacySignBuffer' in jsonDict:
            if jsonDict['LegacySignBuffer'] != None:
                self.legacySignBuffer = bytes.fromhex(
                    jsonDict['LegacySignBuffer']
                )
        if 'Outdated' in jsonDict:
            if jsonDict['Outdated'] != None:
                self.isLegacy = jsonDict['Outdated']
        self.signedOrNeedToSign = jsonDict['Signed'] if 'Signed' in jsonDict else False

    def Dumps(self) -> dict:
        return {
            'Signer': self.signer if self.isLegacy == False and self.signedOrNeedToSign == True else None,
            'Prove': self.prove if self.prove != '' else None,
            'Signature': self.signature.hex() if self.isLegacy == False and self.signedOrNeedToSign == True else None,
            'LegacySignBuffer': self.legacySignBuffer.hex() if self.isLegacy == True and self.signedOrNeedToSign == True else None,
            'Signed': self.signedOrNeedToSign,
            'Outdated': self.isLegacy if self.signedOrNeedToSign == True else None
        }



from io import BytesIO
from ..General.GeneralClass import GeneralClass
from ..utils.getString import getString


class CreateConstantString(GeneralClass):
    def __init__(self) -> None:
        self.operationName: str = 'CreateConstantString'
        self.operationNumber: int = 1
        self.constantString: str = ''

    def Marshal(self, writer: BytesIO) -> None:
        writer.write(self.constantString.encode(encoding='utf-8') + b'\x00')

    def UnMarshal(self, buffer: BytesIO) -> None:
        self.constantString = getString(buffer)



from ..Operation.CreateConstantString import CreateConstantString
from ..Operation.PlaceBlockWithBlockStates import PlaceBlockWithBlockStates
from ..Operation.AddInt16ZValue0 import AddInt16ZValue0
from ..Operation.PlaceBlock import PlaceBlock
from ..Operation.AddZValue0 import AddZValue0
from ..Operation.NOP import NOP
from ..Operation.AddInt32ZValue0 import AddInt32ZValue0
from ..Operation.PlaceBlockWithBlockStatesDeprecated import PlaceBlockWithBlockStatesDeprecated
from ..Operation.AddXValue import AddXValue
from ..Operation.SubtractXValue import SubtractXValue
from ..Operation.AddYValue import AddYValue
from ..Operation.SubtractYValue import SubtractYValue
from ..Operation.AddZValue import AddZValue
from ..Operation.SubtractZValue import SubtractZValue
from ..Operation.AddInt16XValue import AddInt16XValue
from ..Operation.AddInt32XValue import AddInt32XValue
from ..Operation.AddInt16YValue import AddInt16YValue
from ..Operation.AddInt32YValue import AddInt32YValue
from ..Operation.AddInt16XValue import AddInt16XValue
from ..Operation.AddInt32XValue import AddInt32XValue
from ..Operation.AddInt16YValue import AddInt16YValue
from ..Operation.AddInt32YValue import AddInt32YValue
from ..Operation.AddInt16ZValue import AddInt16ZValue
from ..Operation.AddInt32ZValue import AddInt32ZValue
from ..Operation.SetCommandBlockData import SetCommandBlockData
from ..Operation.PlaceBlockWithCommandBlockData import PlaceBlockWithCommandBlockData
from ..Operation.AddInt8XValue import AddInt8XValue
from ..Operation.AddInt8YValue import AddInt8YValue
from ..Operation.AddInt8ZValue import AddInt8ZValue
from ..Operation.UseRuntimeIDPool import UseRuntimeIDPool
from ..Operation.PlaceRuntimeBlock import PlaceRuntimeBlock
from ..Operation.PlaceBlockWithRuntimeId import PlaceBlockWithRuntimeId
from ..Operation.PlaceRuntimeBlockWithCommandBlockData import PlaceRuntimeBlockWithCommandBlockData
from ..Operation.PlaceRuntimeBlockWithCommandBlockDataAndUint32RuntimeID import PlaceRuntimeBlockWithCommandBlockDataAndUint32RuntimeID
from ..Operation.PlaceCommandBlockWithCommandBlockData import PlaceCommandBlockWithCommandBlockData
from ..Operation.PlaceRuntimeBlockWithChestData import PlaceRuntimeBlockWithChestData
from ..Operation.PlaceRuntimeBlockWithChestDataAndUint32RuntimeID import PlaceRuntimeBlockWithChestDataAndUint32RuntimeID
from ..Operation.AssignDebugData import AssignDebugData
from ..Operation.PlaceBlockWithChestData import PlaceBlockWithChestData
from ..Operation.PlaceBlockWithNBTData import PlaceBlockWithNBTData

from ..Operation.structOfChest import ChestSlot, ChestData



from ..General.GeneralClass import GeneralClass
from ..General.Pool import GetBDXCommandPool
from ..utils.getString import getByte, getString
from .Signature import Signature
from .ErrorClassDefine import headerError
from .ErrorClassDefine import readError, unknownOperationError
from brotli import compress, decompress
from io import BytesIO
from copy import deepcopy


class BDX(GeneralClass):
    def __init__(self) -> None:
        """
        `AuthorName: str`
            The author of this BDX file
                Note: The default value is "TriM-Organization/BDXConverter"
        `BDXContent: list[GeneralClass]`
            The valid contents of the BDX file
        `Signature: Signature`
            The signature data of the BDX file
        """
        self.AuthorName: str = 'TriM-Organization/BDXConverter'
        self.BDXContent: list[GeneralClass] = []
        self.Signature: Signature = Signature()

    def Marshal(self, writer: BytesIO) -> None:
        newWriter = BytesIO(
            b'BDX\x00'+self.AuthorName.encode(encoding='utf-8')+b'\x00')
        newWriter.seek(0, 2)
        # inside header with author's name
        for i in self.BDXContent:
            newWriter.write(i.operationNumber.to_bytes(
                length=1, byteorder='big', signed=False))
            i.Marshal(newWriter)
        # valid contents
        if self.Signature.signedOrNeedToSign == True and self.Signature.isLegacy == False:
            self.Signature.BDXContentWithInsideHeader = newWriter
            self.Signature.Marshal(newWriter)
        else:
            newWriter.write(b'XE')
        # signature
        writer.write(b'BD@'+compress(newWriter.getvalue()))
        # write BDX datas

    def UnMarshal(self, binaryDatas: bytes) -> None:
        if binaryDatas[0:3] != b'BD@':
            raise headerError(binaryDatas[:3])
        # check outside header
        reader = BytesIO(decompress(binaryDatas[3:]))
        # get reader to read valid contents
        insideHeader = getByte(reader, 3)
        if insideHeader != b'BDX':
            raise headerError(insideHeader)
        # check inside header
        firstOperation = getByte(reader, 1)
        if firstOperation != b'\x00':
            raise unknownOperationError(firstOperation[0], reader.seek(-1, 1))
        self.AuthorName = getString(reader)
        # get author's name
        BDXCommandPool: dict[int, GeneralClass] = GetBDXCommandPool()
        # get bdx command(operation) pool
        while True:
            commandId = getByte(reader, 1)[0]
            if commandId == 88:
                break
            elif commandId in BDXCommandPool:
                struct: GeneralClass = deepcopy(BDXCommandPool[commandId])
                # get struct(operation) from the pool
                errorType = 0
                # prepare
                try:
                    struct.UnMarshal(reader)
                except EOFError:
                    errorType = 1
                except:
                    errorType = 2
                # unmarshal bytes into the struct
                if errorType == 1:
                    raise EOFError
                elif errorType == 2:
                    raise readError(reader.seek(0, 1))
                # if meet error
                self.BDXContent.append(struct)
                # submit single datas
        # read datas from reader
        self.Signature.UnMarshal(reader)
        if self.Signature.isLegacy == False:
            reader.truncate(reader.seek(-1, 1))
            self.Signature.BDXContentWithInsideHeader = reader
            self.Signature.verifySignature()
        # signature

    def Loads(self, jsonDict: dict) -> None:
        BDXCommandPool: dict[int, GeneralClass] = GetBDXCommandPool()

        self.AuthorName = jsonDict['AuthorName'] if 'AuthorName' in jsonDict else ''
        if 'Signature' in jsonDict:
            self.Signature.Loads(jsonDict['Signature'])
        if 'BDXContent' in jsonDict:
            tmp: list[dict] = jsonDict['BDXContent']
            for i in tmp:
                if not 'operationNumber' in i:
                    continue
                if not 'operationDatas' in i:
                    continue
                commandId: int = i['operationNumber']
                struct: GeneralClass = deepcopy(BDXCommandPool[commandId])
                struct.Loads(i['operationDatas'])
                self.BDXContent.append(struct)

    def Dumps(self) -> dict:
        return {
            'AuthorName': self.AuthorName,
            'BDXContent': [i.Dumps() for i in self.BDXContent],
            'Signature': self.Signature.Dumps()
        }
